#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

/**** sample: TMP107 serial setting: begin ****/
#define SERIAL_BAUDRATE B57600
/**** sample: TMP107 serial setting: end ****/
#define BUF_SIZE 2048

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

/**** sample: TMP107 command: begin ****/
static unsigned char cmd_init[] = {0x55, 0x95, 0x05}; /* set TMP107: init */
static unsigned char cmd_reset[] = {0x55, 0x5d}; /* set TMP107: reset */
static unsigned char cmd_conf[] = {0x55, 0x01, 0xa1, 0x00, 0xa0}; /* set TMP107: configuration */
static unsigned char cmd_read_i[] = {0x55, 0x02, 0xa0}; /* set TMP107: individual read */
static unsigned char cmd_read_g[] = {0x55, 0x03, 0xa0}; /* set TMP107: global read */
/**** sample: TMP107 command: end ****/

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

/**** sample: char(HEX) to int: begin ****/ 
int ch2i(unsigned char c) {
	int ret;
	
	ret = 0;

	if(c>='0' && c<='9') ret = (unsigned int)c - (unsigned int)'0';	
	else if(c>='a' && c <='f') ret = (unsigned int)c - (unsigned int)'a' + 10;	
	else if(c>='A' && c <='F') ret = (unsigned int)c - (unsigned int)'A' + 10;
	else ret = -1;

	return ret;	
}
/**** sample: char(HEX) to int: end ****/ 

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};

/**** sample: variable: begin ****/
	ssize_t ret;
       unsigned char buf[BUF_SIZE];
	int i;
	FILE *fp;
/**** sample: variable: end ****/

/**** sample: variable for senseor data: begin ****/
	int num_tmp;
	float tmp[32];
	unsigned char abs_tmp[2];
	int addr, reg;
/**** sample: for senseor data: end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample: TMP107 initialize: begin ****/
	ret = write(serial_fd, cmd_init, sizeof(cmd_init));
	if(ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: ");
	for(i = 0; i < (int)sizeof(cmd_init); i++) printf("0x%02x ", cmd_init[i]);
	printf("\n");

	sleep(1);

	ret = read(serial_fd, buf, BUF_SIZE);
	if(ret < 0) exitfail_errno("read");
	printf("-------- Read data: ");
	for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
	printf("\n");
	num_tmp = ret - 3;
	printf("-------- Number of TMP107: %d\n", num_tmp);
/**** sample: TMP107 initialize: end ****/

/**** sample: TMP107 reset: begin ****/
	ret = write(serial_fd, cmd_reset, sizeof(cmd_reset));
	if(ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: ");
	for(i = 0; i < (int)sizeof(cmd_reset); i++) printf("0x%02x ", cmd_reset[i]);
	printf("\n");

	sleep(1);

	ret = read(serial_fd, buf, BUF_SIZE);
	if(ret < 0) exitfail_errno("read");
	printf("-------- Read data: ");
	for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
	printf("\n");
/**** sample: TMP107 reset: end ****/

/**** sample: TMP107 configuration: begin ****/
	ret = write(serial_fd, cmd_conf, sizeof(cmd_conf));
	if(ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: ");
	for(i = 0; i < (int)sizeof(cmd_conf); i++) printf("0x%02x ", cmd_conf[i]);
	printf("\n");

	sleep(1);

	ret = read(serial_fd, buf, BUF_SIZE);
	if(ret < 0) exitfail_errno("read");
	printf("-------- Read data: ");
	for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
	printf("\n");
/**** sample: TMP107 configuration: end ****/

/**** sample: TMP107 individual read registers: begin ****/
/*
	printf("\n****************************************************\n");
	printf("-------- Command: Individual Read\n");

	for(addr = 0; addr < num_tmp; addr++) {
		cmd_read_i[1] = 0x02 + (addr << 3);
		printf("-------- Address: 0x%02x\n", addr);
		for(reg = 0; reg < 16; reg++) { 
			cmd_read_i[2] = 0xa0 + reg;
			ret = write(serial_fd, cmd_read_i, sizeof(cmd_read_i));
			if(ret < 0) exitfail_errno("write");

			sleep(1);

			ret = read(serial_fd, buf, BUF_SIZE);
			if(ret < 0) exitfail_errno("read");
			printf("-------- Read data: ");
			for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
			printf("\n");
		}
	}
*/
/**** sample: TMP107 individual read registers: end ****/

/**** sample: TMP107 global read registers: begin ****/
/*
	printf("\n****************************************************\n");
	printf("-------- Command: Global Read\n");

	addr = 0;
	cmd_read_g[1] = 0x03 + ((num_tmp - 1) << 3);
	printf("-------- Start Address: 0x%02x\n", cmd_read_g[1]);
	for(reg = 0; reg < 16; reg++) { 
		cmd_read_g[2] = 0xa0 + reg;
		printf("-------- Register: 0x%02x\n", reg);
		ret = write(serial_fd, cmd_read_g, sizeof(cmd_read_g));
		if(ret < 0) exitfail_errno("write");

		sleep(1);

		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");
		printf("-------- Read data: ");
		for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
		printf("\n");
	}
*/
/**** sample: TMP107 global read registers: end ****/

/**** sample: TMP107 global read temperature register: begin ****/
	printf("\n****************************************************\n");
	printf("-------- Command: Global Read(Temperature)\n");

	addr = 0;
	cmd_read_g[1] = 0x03 + ((num_tmp - 1) << 3);
	printf("-------- Start Address: 0x%02x\n", cmd_read_g[1]);

	reg = 0x00;
	cmd_read_g[2] = 0xa0 + reg;
	printf("-------- Temperature Register: 0x%02x\n", reg);

	ret = write(serial_fd, cmd_read_g, sizeof(cmd_read_g));
	if(ret < 0) exitfail_errno("write");

	sleep(1);

	ret = read(serial_fd, buf, BUF_SIZE);
	if(ret < 0) exitfail_errno("read");
	printf("-------- Read data: ");
	for(i = 0; i < ret; i++) printf("0x%02x ", buf[i]);
	printf("\n");

	for(i = 0; i < num_tmp; i++) {
		if((buf[ret - 1 - (i * 2)] >> 7) == 0) {
			abs_tmp[0] = (buf[ret - 1 - (i * 2)] >> 2) & 0x3f;
			abs_tmp[1] = ((buf[ret - 1 - (i * 2)] << 6) & 0xc0) 
					+ ((buf[ret - 2 - (i * 2)] >> 2) & 0x3f);
			tmp[i] 
				= (float)(abs_tmp[0] * 256 + abs_tmp[1]) * 0.015625;
		}
		else {
			abs_tmp[0] = (~buf[ret - 1 - (i * 2)] >> 2) & 0x3f;
			abs_tmp[1] = ((~buf[ret - 1 - (i * 2)] << 6) & 0xc0) 
					+ ((~buf[ret - 2 - (i * 2)] >> 2) & 0x3f) + 1;
			tmp[i] 
				= (float)(abs_tmp[0] * 256 + abs_tmp[1]) * (-0.015625); 
		}
	}

	printf("-------- Temperature\n");
	for(i = 0; i < num_tmp; i++) {
		printf("Address(0x%02x): %.3f[degC]\n", i, tmp[i]);
	}

/**** sample: sensor data to log file for post to heroku: begin ****/
	fp = fopen("log_tmp.txt", "w");
	for(i = 0; i < num_tmp; i++)
		fprintf(fp, "tmp%d=%.3f&", i, tmp[i]);
	fclose(fp);
/**** sample: sensor data to log file for post to heroku: end ****/


	return EXIT_SUCCESS;
}

